@@ -4,24 +4,19 @@ import enum
 from typing import Sequence, Optional, Dict, TYPE_CHECKING
 from abc import abstractmethod, ABC
 
-from PyQt5 import QtCore, QtGui
-from PyQt5.QtCore import Qt, QRect, QSize
-from PyQt5.QtWidgets import (QMenu, QHBoxLayout, QLabel, QVBoxLayout, QGridLayout, QLineEdit,
-                             QPushButton, QAbstractItemView, QComboBox, QCheckBox,
-                             QToolTip)
-from PyQt5.QtGui import QFont, QStandardItem, QBrush, QPainter, QIcon, QHelpEvent
+from PyQt6 import QtCore, QtGui
+from PyQt6.QtCore import Qt, QRect, QSize
+from PyQt6.QtWidgets import QMenu, QLabel, QVBoxLayout, QGridLayout, QAbstractItemView, QCheckBox, QToolTip
+from PyQt6.QtGui import QFont, QStandardItem, QBrush, QPainter, QIcon, QHelpEvent
 
-from electrum.util import NotEnoughFunds, NoDynamicFeeEstimates
 from electrum.i18n import _
-from electrum.lnchannel import AbstractChannel, PeerState, ChannelBackup, Channel, ChannelState, ChanCloseOption
+from electrum.lnchannel import AbstractChannel, ChannelBackup, Channel, ChanCloseOption
 from electrum.wallet import Abstract_Wallet
-from electrum.lnutil import LOCAL, REMOTE, format_short_channel_id
+from electrum.lnutil import LOCAL, REMOTE
 from electrum.lnworker import LNWallet
 from electrum.gui import messages
 
-from .util import (WindowModalDialog, Buttons, OkButton, CancelButton,
-                   EnterButton, WaitingDialog, MONOSPACE_FONT, ColorScheme)
-from .amountedit import BTCAmountEdit, FreezableLineEdit
+from .util import WindowModalDialog, Buttons, OkButton, EnterButton, WaitingDialog, MONOSPACE_FONT, ColorScheme
 from .util import read_QIcon, font_height
 from .my_treeview import MyTreeView
 
@@ -29,7 +24,7 @@ if TYPE_CHECKING:
     from .main_window import ElectrumWindow
 
 
-ROLE_CHANNEL_ID = Qt.UserRole
+ROLE_CHANNEL_ID = Qt.ItemDataRole.UserRole
 
 
 class ChannelsList(MyTreeView):
@@ -73,7 +68,7 @@ class ChannelsList(MyTreeView):
             stretch_column=self.Columns.NODE_ALIAS,
         )
         self.setModel(QtGui.QStandardItemModel(self))
-        self.setSelectionMode(QAbstractItemView.ExtendedSelection)
+        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
         self.gossip_db_loaded.connect(self.on_gossip_db)
         self.update_rows.connect(self.do_update_rows)
         self.update_single_row.connect(self.do_update_single_row)
@@ -103,7 +98,7 @@ class ChannelsList(MyTreeView):
             labels[subject] = label
         status = chan.get_state_for_GUI()
         closed = chan.is_closed()
-        node_alias = self.lnworker.get_node_alias(chan.node_id) or chan.node_id.hex()
+        node_alias = self.lnworker.lnpeermgr.get_node_alias(chan.node_id) or chan.node_id.hex()
         capacity_str = self.main_window.format_amount(chan.get_capacity(), whitespaces=True)
         return {
             self.Columns.SHORT_CHANID: chan.short_id_for_GUI(),
@@ -117,10 +112,7 @@ class ChannelsList(MyTreeView):
         }
 
     def on_channel_closed(self, txid):
-        self.main_window.show_error('Channel closed' + '\n' + txid)
-
-    def on_request_sent(self, b):
-        self.main_window.show_message(_('Request sent'))
+        self.main_window.show_message('Channel closed' + '\n' + txid)
 
     def on_failure(self, exc_info):
         type_, e, tb = exc_info
@@ -130,20 +122,24 @@ class ChannelsList(MyTreeView):
     def close_channel(self, channel_id):
         self.is_force_close = False
         msg = _('Cooperative close?')
-        msg += '\n' + _(messages.MSG_COOPERATIVE_CLOSE)
+        msg += '\n\n' + messages.MSG_COOPERATIVE_CLOSE
         if not self.main_window.question(msg):
             return
         coro = self.lnworker.close_channel(channel_id)
         on_success = self.on_channel_closed
+
         def task():
             return self.network.run_from_another_thread(coro)
-        WaitingDialog(self, 'please wait..', task, on_success, self.on_failure)
+
+        WaitingDialog(self, _('Please wait...'), task, on_success, self.on_failure)
 
     def force_close(self, channel_id):
         self.save_backup = True
         backup_cb = QCheckBox('Create a backup now', checked=True)
-        def on_checked(b):
-            self.save_backup = bool(b)
+
+        def on_checked(_x):
+            self.save_backup = backup_cb.isChecked()
+
         backup_cb.stateChanged.connect(on_checked)
         chan = self.lnworker.channels[channel_id]
         to_self_delay = chan.config[REMOTE].to_self_delay
@@ -158,10 +154,12 @@ class ChannelsList(MyTreeView):
         if self.save_backup:
             if not self.main_window.backup_wallet():
                 return
+
         def task():
             coro = self.lnworker.force_close_channel(channel_id)
             return self.network.run_from_another_thread(coro)
-        WaitingDialog(self, 'please wait..', task, self.on_channel_closed, self.on_failure)
+
+        WaitingDialog(self, _('Please wait...'), task, self.on_channel_closed, self.on_failure)
 
     def remove_channel(self, channel_id):
         if self.main_window.question(_('Are you sure you want to delete this channel? This will purge associated transactions from your wallet history.')):
@@ -172,30 +170,32 @@ class ChannelsList(MyTreeView):
             self.lnworker.remove_channel_backup(channel_id)
 
     def export_channel_backup(self, channel_id):
-        msg = ' '.join([
-            _("Channel backups can be imported in another instance of the same wallet."),
-            _("In the Electrum mobile app, use the 'Send' button to scan this QR code."),
-            '\n\n',
-            _("Please note that channel backups cannot be used to restore your channels."),
-            _("If you lose your wallet file, the only thing you can do with a backup is to request your channel to be closed, so that your funds will be sent on-chain."),
-        ])
+        msg = messages.MSG_LN_EXPLAIN_SCB_BACKUPS
         data = self.lnworker.export_channel_backup(channel_id)
         self.main_window.show_qrcode(data, 'channel backup', help_text=msg,
                                      show_copy_text_btn=True)
 
     def request_force_close(self, channel_id):
         msg = _('Request force-close from remote peer?')
-        msg += '\n' + _(messages.MSG_REQUEST_FORCE_CLOSE)
+        msg += '\n\n' + messages.MSG_REQUEST_FORCE_CLOSE
         if not self.main_window.question(msg):
             return
+
         def task():
             coro = self.lnworker.request_force_close(channel_id)
             return self.network.run_from_another_thread(coro)
-        WaitingDialog(self, 'please wait..', task, self.on_request_sent, self.on_failure)
 
-    def freeze_channel_for_sending(self, chan, b):
+        def on_done(b):
+            self.main_window.show_message(_('Request scheduled'))
+
+        WaitingDialog(self, _('Please wait...'), task, on_done, self.on_failure)
+
+    def set_frozen(self, chan, *, for_sending, value):
         if not self.lnworker.uses_trampoline() or self.lnworker.is_trampoline_peer(chan.node_id):
-            chan.set_frozen_for_sending(b)
+            if for_sending:
+                chan.set_frozen_for_sending(value)
+            else:
+                chan.set_frozen_for_receiving(value)
         else:
             msg = messages.MSG_NON_TRAMPOLINE_CHANNEL_FROZEN_WITHOUT_GOSSIP
             self.main_window.show_warning(msg, title=_('Channel is frozen for sending'))
@@ -207,8 +207,8 @@ class ChannelsList(MyTreeView):
             idx2 = selected[1]
             channel_id1 = idx1.sibling(idx1.row(), self.Columns.NODE_ALIAS).data(ROLE_CHANNEL_ID)
             channel_id2 = idx2.sibling(idx2.row(), self.Columns.NODE_ALIAS).data(ROLE_CHANNEL_ID)
-            chan1 = self.lnworker.channels.get(channel_id1)
-            chan2 = self.lnworker.channels.get(channel_id2)
+            chan1 = self.lnworker.get_channel_by_id(channel_id1)
+            chan2 = self.lnworker.get_channel_by_id(channel_id2)
             if chan1 and chan2 and (not self.lnworker.uses_trampoline() or chan1.node_id != chan2.node_id):
                 return chan1, chan2
         return None, None
@@ -230,13 +230,13 @@ class ChannelsList(MyTreeView):
         menu.setSeparatorsCollapsible(True)  # consecutive separators are merged together
         selected = self.selected_in_column(self.Columns.NODE_ALIAS)
         if not selected:
-            menu.exec_(self.viewport().mapToGlobal(position))
+            menu.exec(self.viewport().mapToGlobal(position))
             return
         if len(selected) == 2:
             chan1, chan2 = self.get_rebalance_pair()
             if chan1 and chan2:
                 menu.addAction(_("Rebalance channels"), lambda: self.main_window.rebalance_dialog(chan1, chan2))
-                menu.exec_(self.viewport().mapToGlobal(position))
+                menu.exec(self.viewport().mapToGlobal(position))
             return
         elif len(selected) > 2:
             return
@@ -258,13 +258,13 @@ class ChannelsList(MyTreeView):
         if not chan.is_backup() and not chan.is_closed():
             fm = menu.addMenu(_("Freeze"))
             if not chan.is_frozen_for_sending():
-                fm.addAction(_("Freeze for sending"), lambda: self.freeze_channel_for_sending(chan, True))
+                fm.addAction(_("Freeze for sending"), lambda: self.set_frozen(chan, for_sending=True, value=True))
             else:
-                fm.addAction(_("Unfreeze for sending"), lambda: self.freeze_channel_for_sending(chan, False))
+                fm.addAction(_("Unfreeze for sending"), lambda: self.set_frozen(chan, for_sending=True, value=False))
             if not chan.is_frozen_for_receiving():
-                fm.addAction(_("Freeze for receiving"), lambda: chan.set_frozen_for_receiving(True))
+                fm.addAction(_("Freeze for receiving"), lambda: self.set_frozen(chan, for_sending=False, value=True))
             else:
-                fm.addAction(_("Unfreeze for receiving"), lambda: chan.set_frozen_for_receiving(False))
+                fm.addAction(_("Unfreeze for receiving"), lambda: self.set_frozen(chan, for_sending=False, value=False))
         if close_opts := chan.get_close_options():
             cm = menu.addMenu(_("Close"))
             if ChanCloseOption.COOP_CLOSE in close_opts:
@@ -281,7 +281,7 @@ class ChannelsList(MyTreeView):
                 menu.addAction(_("Delete"), lambda: self.remove_channel_backup(channel_id))
             else:
                 menu.addAction(_("Delete"), lambda: self.remove_channel(channel_id))
-        menu.exec_(self.viewport().mapToGlobal(position))
+        menu.exec(self.viewport().mapToGlobal(position))
 
     @QtCore.pyqtSlot(Abstract_Wallet, AbstractChannel)
     def do_update_single_row(self, wallet: Abstract_Wallet, chan: AbstractChannel):
@@ -292,7 +292,7 @@ class ChannelsList(MyTreeView):
             if item.data(ROLE_CHANNEL_ID) != chan.channel_id:
                 continue
             for column, v in self.format_fields(chan).items():
-                self.model().item(row, column).setData(v, QtCore.Qt.DisplayRole)
+                self.model().item(row, column).setData(v, QtCore.Qt.ItemDataRole.DisplayRole)
             items = [self.model().item(row, column) for column in self.Columns]
             self._update_chan_frozen_bg(chan=chan, items=items)
         if wallet.lnworker:
@@ -328,7 +328,8 @@ class ChannelsList(MyTreeView):
             self._update_chan_frozen_bg(chan=chan, items=items)
             self.model().insertRow(0, items)
 
-        self.sortByColumn(self.Columns.SHORT_CHANID, Qt.DescendingOrder)
+        # FIXME sorting by SHORT_CHANID should treat values as tuple, not as string ( 50x1x1 > 8x1x1 )
+        self.sortByColumn(self.Columns.SHORT_CHANID, Qt.SortOrder.DescendingOrder)
 
     def _update_chan_frozen_bg(self, *, chan: AbstractChannel, items: Sequence[QStandardItem]):
         assert self._default_item_bg_brush is not None
@@ -363,8 +364,15 @@ class ChannelsList(MyTreeView):
         menu.addAction(read_QIcon('update.png'), _('Submarine swap'), lambda: self.main_window.run_swap_dialog())
         menu.addSeparator()
         menu.addAction(_("Import channel backup"), lambda: self.main_window.do_process_from_text_channel_backup())
+        # only enable menu if has LN. Or we could selectively enable menu items?
+        #     and maybe add item "main_window.init_lightning_dialog()" when applicable
+        menu.setEnabled(self.wallet.has_lightning())
         self.new_channel_button = EnterButton(_('New Channel'), self.main_window.new_channel_dialog)
-        self.new_channel_button.setEnabled(self.wallet.has_lightning())
+        if not self.wallet.can_have_lightning():
+            self.new_channel_button.setEnabled(False)
+            self.new_channel_button.setToolTip(_("Lightning is not available for this wallet."))
+        else:
+            self.new_channel_button.setToolTip(_("Open a channel to send payments over the Lightning network."))
         toolbar.insertWidget(2, self.new_channel_button)
         return toolbar
 
@@ -383,7 +391,7 @@ class ChannelsList(MyTreeView):
         h.addWidget(QLabel(capacity), 2, 1)
         vbox.addLayout(h)
         vbox.addLayout(Buttons(OkButton(d)))
-        d.exec_()
+        d.exec()
 
     def set_visibility_of_columns(self):
         def set_visible(col: int, b: bool):
@@ -432,6 +440,7 @@ class ChanFeatNoOnchainBackup(ChannelFeature):
         return read_QIcon("cloud_no")
 
 
+
 class ChannelFeatureIcons:
 
     def __init__(self, features: Sequence['ChannelFeature']):
