@@ -1,26 +1,22 @@
-
 # source: http://stackoverflow.com/questions/2758159/how-to-embed-a-python-interpreter-in-a-pyqt-widget
 
 import sys
 import os
 import re
 import traceback
-import platform
 
-from PyQt5 import QtCore
-from PyQt5 import QtGui
-from PyQt5 import QtWidgets
+from PyQt6 import QtCore, QtGui, QtWidgets
+from PyQt6.QtCore import Qt
 
 from electrum import util
 from electrum.i18n import _
+from electrum.base_crash_reporter import taint_reports_by_console_usage
 
+from .util import MONOSPACE_FONT, font_height
 
-if platform.system() == 'Windows':
-    MONOSPACE_FONT = 'Lucida Console'
-elif platform.system() == 'Darwin':
-    MONOSPACE_FONT = 'Monaco'
-else:
-    MONOSPACE_FONT = 'monospace'
+# sys.ps1 and sys.ps2 are only declared if an interpreter is in interactive mode.
+sys.ps1 = '>>> '
+sys.ps2 = '... '
 
 
 class OverlayLabel(QtWidgets.QLabel):
@@ -35,11 +31,11 @@ class OverlayLabel(QtWidgets.QLabel):
     '''
     def __init__(self, text, parent):
         super().__init__(text, parent)
-        self.setMinimumHeight(150)
+        self.setMinimumHeight(max(150, 10 * font_height()))
         self.setGeometry(0, 0, self.width(), self.height())
         self.setStyleSheet(self.STYLESHEET)
         self.setMargin(0)
-        parent.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
+        parent.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
         self.setWordWrap(True)
 
     def mousePressEvent(self, e):
@@ -51,19 +47,23 @@ class OverlayLabel(QtWidgets.QLabel):
 
 
 class Console(QtWidgets.QPlainTextEdit):
-    def __init__(self, prompt='>> ', startup_message='', parent=None):
+    DEFAULT_FONT_SIZE = 10
+    MIN_FONT_SIZE = 6
+    MAX_FONT_SIZE = 32
+
+    def __init__(self, parent=None):
         QtWidgets.QPlainTextEdit.__init__(self, parent)
 
-        self.prompt = prompt
         self.history = []
         self.namespace = {}
         self.construct = []
+        self.font_size = self.DEFAULT_FONT_SIZE
 
         self.setGeometry(50, 75, 600, 400)
-        self.setWordWrapMode(QtGui.QTextOption.WrapAnywhere)
+        self.setWordWrapMode(QtGui.QTextOption.WrapMode.WrapAnywhere)
         self.setUndoRedoEnabled(False)
-        self.document().setDefaultFont(QtGui.QFont(MONOSPACE_FONT, 10, QtGui.QFont.Normal))
-        self.showMessage(startup_message)
+        self.setFont(QtGui.QFont(MONOSPACE_FONT, self.font_size, QtGui.QFont.Weight.Normal))
+        self.newPrompt("")  # make sure there is always a prompt, even before first server.banner
 
         self.updateNamespace({'run':self.run_script})
         self.set_json(False)
@@ -76,6 +76,11 @@ class Console(QtWidgets.QPlainTextEdit):
         )
         self.messageOverlay = OverlayLabel(warning_text, self)
 
+    def set_font_size(self, size: int):
+        size = max(self.MIN_FONT_SIZE, min(self.MAX_FONT_SIZE, size))
+        self.font_size = size
+        self.setFont(QtGui.QFont(MONOSPACE_FONT, self.font_size, QtGui.QFont.Weight.Normal))
+
     def resizeEvent(self, e):
         super().resizeEvent(e)
         vertical_scrollbar_width = self.verticalScrollBar().width() * self.verticalScrollBar().isVisible()
@@ -88,39 +93,44 @@ class Console(QtWidgets.QPlainTextEdit):
         with open(filename) as f:
             script = f.read()
 
-        # eval is generally considered bad practice. use it wisely!
-        result = eval(script, self.namespace, self.namespace)
-
-
+        self._exec_command(script)
 
     def updateNamespace(self, namespace):
         self.namespace.update(namespace)
 
     def showMessage(self, message):
+        curr_line = self.getCommand(strip=False)
         self.appendPlainText(message)
-        self.newPrompt()
+        self.newPrompt(curr_line)
 
     def clear(self):
+        curr_line = self.getCommand()
         self.setPlainText('')
-        self.newPrompt()
+        self.newPrompt(curr_line)
 
-    def newPrompt(self):
+    def keyboard_interrupt(self):
+        self.construct = []
+        self.appendPlainText('KeyboardInterrupt')
+        self.newPrompt('')
+
+    def newPrompt(self, curr_line):
         if self.construct:
-            prompt = '.' * len(self.prompt)
+            prompt = sys.ps2 + curr_line
         else:
-            prompt = self.prompt
+            prompt = sys.ps1 + curr_line
 
         self.completions_pos = self.textCursor().position()
         self.completions_visible = False
 
         self.appendPlainText(prompt)
-        self.moveCursor(QtGui.QTextCursor.End)
+        self.moveCursor(QtGui.QTextCursor.MoveOperation.End)
 
-    def getCommand(self):
+    def getCommand(self, *, strip=True):
         doc = self.document()
         curr_line = doc.findBlockByLineNumber(doc.lineCount() - 1).text()
-        curr_line = curr_line.rstrip()
-        curr_line = curr_line[len(self.prompt):]
+        if strip:
+            curr_line = curr_line.rstrip()
+        curr_line = curr_line[len(sys.ps1):]
         return curr_line
 
     def setCommand(self, command):
@@ -129,13 +139,13 @@ class Console(QtWidgets.QPlainTextEdit):
 
         doc = self.document()
         curr_line = doc.findBlockByLineNumber(doc.lineCount() - 1).text()
-        self.moveCursor(QtGui.QTextCursor.End)
-        for i in range(len(curr_line) - len(self.prompt)):
-            self.moveCursor(QtGui.QTextCursor.Left, QtGui.QTextCursor.KeepAnchor)
+        self.moveCursor(QtGui.QTextCursor.MoveOperation.End)
+        for i in range(len(curr_line) - len(sys.ps1)):
+            self.moveCursor(QtGui.QTextCursor.MoveOperation.Left, QtGui.QTextCursor.MoveMode.KeepAnchor)
 
         self.textCursor().removeSelectedText()
         self.textCursor().insertText(command)
-        self.moveCursor(QtGui.QTextCursor.End)
+        self.moveCursor(QtGui.QTextCursor.MoveOperation.End)
 
     def show_completions(self, completions):
         if self.completions_visible:
@@ -151,7 +161,7 @@ class Console(QtWidgets.QPlainTextEdit):
         c.insertText(t)
         self.completions_end = c.position()
 
-        self.moveCursor(QtGui.QTextCursor.End)
+        self.moveCursor(QtGui.QTextCursor.MoveOperation.End)
         self.completions_visible = True
 
     def hide_completions(self):
@@ -162,14 +172,13 @@ class Console(QtWidgets.QPlainTextEdit):
         l = self.completions_end - self.completions_pos
         for x in range(l): c.deleteChar()
 
-        self.moveCursor(QtGui.QTextCursor.End)
+        self.moveCursor(QtGui.QTextCursor.MoveOperation.End)
         self.completions_visible = False
 
     def getConstruct(self, command):
         if self.construct:
-            prev_command = self.construct[-1]
             self.construct.append(command)
-            if not prev_command and not command:
+            if not command:
                 ret_val = '\n'.join(self.construct)
                 self.construct = []
                 return ret_val
@@ -182,17 +191,13 @@ class Console(QtWidgets.QPlainTextEdit):
             else:
                 return command
 
-    def getHistory(self):
-        return self.history
-
-    def setHisory(self, history):
-        self.history = history
-
     def addToHistory(self, command):
-        if command[0:1] == ' ':
+        if not self.construct and command[0:1] == ' ':
             return
 
         if command and (not self.history or self.history[-1] != command):
+            while len(self.history) >= 50:
+                self.history.remove(self.history[0])
             self.history.append(command)
         self.history_index = len(self.history)
 
@@ -212,106 +217,119 @@ class Console(QtWidgets.QPlainTextEdit):
 
     def getCursorPosition(self):
         c = self.textCursor()
-        return c.position() - c.block().position() - len(self.prompt)
+        return c.position() - c.block().position() - len(sys.ps1)
 
     def setCursorPosition(self, position):
-        self.moveCursor(QtGui.QTextCursor.StartOfLine)
-        for i in range(len(self.prompt) + position):
-            self.moveCursor(QtGui.QTextCursor.Right)
+        self.moveCursor(QtGui.QTextCursor.MoveOperation.StartOfLine)
+        for i in range(len(sys.ps1) + position):
+            self.moveCursor(QtGui.QTextCursor.MoveOperation.Right)
 
-    def register_command(self, c, func):
-        methods = { c: func}
-        self.updateNamespace(methods)
-
-
-    def runCommand(self):
+    def run_command(self):
         command = self.getCommand()
         self.addToHistory(command)
 
         command = self.getConstruct(command)
 
         if command:
-            tmp_stdout = sys.stdout
-
-            class stdoutProxy():
-                def __init__(self, write_func):
-                    self.write_func = write_func
-                    self.skip = False
-
-                def flush(self):
-                    pass
-
-                def write(self, text):
-                    if not self.skip:
-                        stripped_text = text.rstrip('\n')
-                        self.write_func(stripped_text)
-                        QtCore.QCoreApplication.processEvents()
-                    self.skip = not self.skip
-
-            if type(self.namespace.get(command)) == type(lambda:None):
-                self.appendPlainText("'{}' is a function. Type '{}()' to use it in the Python console."
-                                     .format(command, command))
-                self.newPrompt()
-                return
-
-            sys.stdout = stdoutProxy(self.appendPlainText)
-            try:
-                try:
-                    # eval is generally considered bad practice. use it wisely!
-                    result = eval(command, self.namespace, self.namespace)
-                    if result != None:
-                        if self.is_json:
-                            util.print_msg(util.json_encode(result))
-                        else:
-                            self.appendPlainText(repr(result))
-                except SyntaxError:
-                    # exec is generally considered bad practice. use it wisely!
-                    exec(command, self.namespace, self.namespace)
-            except SystemExit:
-                self.close()
-            except BaseException:
-                traceback_lines = traceback.format_exc().split('\n')
-                # Remove traceback mentioning this file, and a linebreak
-                for i in (3,2,1,-1):
-                    traceback_lines.pop(i)
-                self.appendPlainText('\n'.join(traceback_lines))
-            sys.stdout = tmp_stdout
-        self.newPrompt()
+            self._exec_command(command)
+        self.newPrompt('')
         self.set_json(False)
 
+    def _exec_command(self, command):
+        tmp_stdout = sys.stdout
+        taint_reports_by_console_usage()
+
+        class StdoutProxy:
+            def __init__(self, write_func):
+                self.write_func = write_func
+                self.skip = False
+
+            def flush(self):
+                pass
+
+            def write(self, text):
+                if not self.skip:
+                    stripped_text = text.rstrip('\n')
+                    self.write_func(stripped_text)
+                    QtCore.QCoreApplication.processEvents()
+                self.skip = not self.skip
+
+        if type(self.namespace.get(command)) == type(lambda: None):
+            self.appendPlainText("'{}' is a function. Type '{}()' to use it in the Python console."
+                                 .format(command, command))
+            return
+
+        sys.stdout = StdoutProxy(self.appendPlainText)
+        try:
+            try:
+                # eval is generally considered bad practice. use it wisely!
+                result = eval(command, self.namespace, self.namespace)
+                if result is not None:
+                    if self.is_json:
+                        util.print_msg(util.json_encode(result))
+                    else:
+                        self.appendPlainText(repr(result))
+            except SyntaxError:
+                # exec is generally considered bad practice. use it wisely!
+                exec(command, self.namespace, self.namespace)
+        except SystemExit:
+            self.close()
+        except BaseException as e:
+            te = traceback.TracebackException.from_exception(e)
+            # rm part of traceback mentioning this file.
+            # (note: we rm stack items before converting to str, instead of removing lines from the str,
+            #        as this is more reliable. The latter would differ whether the traceback has source text lines,
+            #        which is not always the case.)
+            te.stack = traceback.StackSummary.from_list(te.stack[1:])
+            tb_str = "".join(te.format())
+            # rm last linebreak:
+            if tb_str.endswith("\n"):
+                tb_str = tb_str[:-1]
+            self.appendPlainText(tb_str)
+        sys.stdout = tmp_stdout
 
     def keyPressEvent(self, event):
-        if event.key() == QtCore.Qt.Key_Tab:
+        if event.key() == Qt.Key.Key_Tab:
             self.completions()
             return
 
         self.hide_completions()
 
-        if event.key() in (QtCore.Qt.Key_Enter, QtCore.Qt.Key_Return):
-            self.runCommand()
+        if event.key() in (Qt.Key.Key_Enter, Qt.Key.Key_Return):
+            self.run_command()
             return
-        if event.key() == QtCore.Qt.Key_Home:
+        if event.key() == Qt.Key.Key_Home:
             self.setCursorPosition(0)
             return
-        if event.key() == QtCore.Qt.Key_PageUp:
+        if event.key() == Qt.Key.Key_PageUp:
             return
-        elif event.key() in (QtCore.Qt.Key_Left, QtCore.Qt.Key_Backspace):
+        elif event.key() in (Qt.Key.Key_Left, Qt.Key.Key_Backspace):
             if self.getCursorPosition() == 0:
                 return
-        elif event.key() == QtCore.Qt.Key_Up:
+        elif event.key() == Qt.Key.Key_Up:
             self.setCommand(self.getPrevHistoryEntry())
             return
-        elif event.key() == QtCore.Qt.Key_Down:
+        elif event.key() == Qt.Key.Key_Down:
             self.setCommand(self.getNextHistoryEntry())
             return
-        elif event.key() == QtCore.Qt.Key_L and event.modifiers() == QtCore.Qt.ControlModifier:
+        elif event.key() == Qt.Key.Key_L and event.modifiers() == Qt.KeyboardModifier.ControlModifier:
             self.clear()
+        elif event.key() == Qt.Key.Key_C and event.modifiers() == Qt.KeyboardModifier.ControlModifier:
+            if not self.textCursor().selectedText():
+                self.keyboard_interrupt()
+        elif event.key() == Qt.Key.Key_Plus and Qt.KeyboardModifier.ControlModifier in event.modifiers():
+            self.set_font_size(self.font_size + 1)
+            return
+        elif event.key() == Qt.Key.Key_Minus and Qt.KeyboardModifier.ControlModifier in event.modifiers():
+            self.set_font_size(self.font_size - 1)
+            return
 
         super(Console, self).keyPressEvent(event)
 
     def completions(self):
         cmd = self.getCommand()
-        lastword = re.split(' |\(|\)',cmd)[-1]
+        # note for regex: new words start after ' ' or '(' or ')'
+        lastword = re.split(r'[ ()]', cmd)[-1]
         beginning = cmd[0:-len(lastword)]
 
         path = lastword.split('.')
@@ -352,17 +370,3 @@ class Console(QtWidgets.QPlainTextEdit):
                 self.setCommand(beginning + p)
             else:
                 self.show_completions(completions)
-
-
-welcome_message = '''
-   ---------------------------------------------------------------
-     Welcome to a primitive Python interpreter.
-   ---------------------------------------------------------------
-'''
-
-if __name__ == '__main__':
-    app = QtWidgets.QApplication(sys.argv)
-    console = Console(startup_message=welcome_message)
-    console.updateNamespace({'myVar1' : app, 'myVar2' : 1234})
-    console.show()
-    sys.exit(app.exec_())
