@@ -36,6 +36,7 @@ from electrum.bitcoin import is_address
 from electrum.transaction import PartialTxInput, PartialTxOutput
 from electrum.lnutil import MIN_FUNDING_SAT
 from electrum.util import profiler
+from electrum.plugin import run_hook
 
 from .util import ColorScheme, MONOSPACE_FONT
 from .my_treeview import MyTreeView, MySortModel
@@ -103,12 +104,11 @@ class UTXOList(MyTreeView):
         self.proxy.setDynamicSortFilter(False)  # temp. disable re-sorting after every change
         utxos = self.wallet.get_utxos()
         self._maybe_reset_coincontrol(utxos)
-        self._utxo_dict = {}
+        self._utxo_dict = dict([(utxo.prevout.to_str(), utxo) for utxo in utxos])
         self.std_model.clear()
         self.update_headers(self.__class__.headers)
         for idx, utxo in enumerate(utxos):
             name = utxo.prevout.to_str()
-            self._utxo_dict[name] = utxo
             labels = [""] * len(self.Columns)
             amount_str = self.main_window.format_amount(
                 utxo.value_sats(), whitespaces=True)
@@ -192,6 +192,7 @@ class UTXOList(MyTreeView):
         return all([utxo.prevout.to_str() in self._spend_set for utxo in coins])
 
     def add_to_coincontrol(self, coins: List[PartialTxInput]):
+        assert all(utxo.prevout.to_str() in self._utxo_dict for utxo in coins) # see issue 10206
         coins = self._filter_frozen_coins(coins)
         for utxo in coins:
             self._spend_set.add(utxo.prevout.to_str())
@@ -255,7 +256,7 @@ class UTXOList(MyTreeView):
     def swap_coins(self, coins):
         #self.clear_coincontrol()
         self.add_to_coincontrol(coins)
-        self.main_window.run_swap_dialog(is_reverse=False, recv_amount_sat='!')
+        self.main_window.run_swap_dialog(is_reverse=False, recv_amount_sat_or_max='!')
         self.clear_coincontrol()
 
     def can_open_channel(self, coins):
@@ -367,6 +368,7 @@ class UTXOList(MyTreeView):
                 act = menu_freeze.addAction(_("Unfreeze Addresses"), lambda: self.main_window.set_frozen_state_of_addresses(addrs, False))
                 act.setToolTip(MSG_FREEZE_ADDRESS)
 
+        run_hook('qt_utxo_menu', menu, coins, self.wallet)
         menu.exec(self.viewport().mapToGlobal(position))
 
     def get_filter_data_from_coordinate(self, row, col):
