@@ -1,45 +1,40 @@
 from copy import deepcopy
-from typing import Optional, Sequence, Tuple, List, Dict, TYPE_CHECKING, Set
-import threading
+from typing import Sequence, Tuple, Dict, TYPE_CHECKING, Set
 
 from .lnutil import SENT, RECEIVED, LOCAL, REMOTE, HTLCOwner, UpdateAddHtlc, Direction, FeeUpdate
-from .util import bh2u, bfh
+from .util import bfh, with_lock
 
 if TYPE_CHECKING:
     from .json_db import StoredDict
 
+LOG_TEMPLATE = {
+    'adds': {},              # "side who offered htlc" -> htlc_id -> htlc
+    'locked_in': {},         # "side who offered htlc" -> action -> htlc_id -> whose ctx -> ctn
+    'settles': {},           # "side who offered htlc" -> action -> htlc_id -> whose ctx -> ctn
+    'fails': {},             # "side who offered htlc" -> action -> htlc_id -> whose ctx -> ctn
+    'fee_updates': {},       # "side who initiated fee update" -> index -> list of FeeUpdates
+    'revack_pending': False,
+    'next_htlc_id': 0,
+    'ctn': -1,               # oldest unrevoked ctx of sub
+}
+
 
 class HTLCManager:
 
-    def __init__(self, log:'StoredDict', *, initial_feerate=None):
+    def __init__(self, log: 'StoredDict', *, initiator=None, initial_feerate=None):
 
         if len(log) == 0:
-            initial = {
-                'adds': {},              # "side who offered htlc" -> htlc_id -> htlc
-                'locked_in': {},         # "side who offered htlc" -> action -> htlc_id -> whose ctx -> ctn
-                'settles': {},           # "side who offered htlc" -> action -> htlc_id -> whose ctx -> ctn
-                'fails': {},             # "side who offered htlc" -> action -> htlc_id -> whose ctx -> ctn
-                'fee_updates': {},       # "side who initiated fee update" -> action -> list of FeeUpdates
-                'revack_pending': False,
-                'next_htlc_id': 0,
-                'ctn': -1,               # oldest unrevoked ctx of sub
-            }
             # note: "htlc_id" keys in dict are str! but due to json_db magic they can *almost* be treated as int...
-            log[LOCAL] = deepcopy(initial)
-            log[REMOTE] = deepcopy(initial)
-            log['unacked_local_updates2'] = {}
-
-        if 'unfulfilled_htlcs' not in log:
-            log['unfulfilled_htlcs'] = {}  # htlc_id -> onion_packet
-        if 'fail_htlc_reasons' not in log:
-            log['fail_htlc_reasons'] = {}  # htlc_id -> error_bytes, failure_message
+            log[LOCAL] = deepcopy(LOG_TEMPLATE)
+            log[REMOTE] = deepcopy(LOG_TEMPLATE)
+            log[LOCAL]['unacked_updates'] = {}
+            log[LOCAL]['was_revoke_last'] = False
 
         # maybe bootstrap fee_updates if initial_feerate was provided
         if initial_feerate is not None:
             assert type(initial_feerate) is int
-            for sub in (LOCAL, REMOTE):
-                if not log[sub]['fee_updates']:
-                    log[sub]['fee_updates'][0] = FeeUpdate(rate=initial_feerate, ctn_local=0, ctn_remote=0)
+            assert initiator in [LOCAL, REMOTE]
+            log[initiator]['fee_updates'][0] = FeeUpdate(rate=initial_feerate, ctn_local=0, ctn_remote=0)
         self.log = log
 
         # We need a lock as many methods of HTLCManager are accessed by both the asyncio thread and the GUI.
@@ -50,12 +45,6 @@ class HTLCManager:
 
         self._init_maybe_active_htlc_ids()
 
-    def with_lock(func):
-        def func_wrapper(self, *args, **kwargs):
-            with self.lock:
-                return func(self, *args, **kwargs)
-        return func_wrapper
-
     @with_lock
     def ctn_latest(self, sub: HTLCOwner) -> int:
         """Return the ctn for the latest (newest that has a valid sig) ctx of sub"""
@@ -166,6 +155,7 @@ class HTLCManager:
     def send_ctx(self) -> None:
         assert self.ctn_latest(REMOTE) == self.ctn_oldest_unrevoked(REMOTE), (self.ctn_latest(REMOTE), self.ctn_oldest_unrevoked(REMOTE))
         self._set_revack_pending(REMOTE, True)
+        self.log[LOCAL]['was_revoke_last'] = False
 
     @with_lock
     def recv_ctx(self) -> None:
@@ -176,6 +166,7 @@ class HTLCManager:
     def send_rev(self) -> None:
         self.log[LOCAL]['ctn'] += 1
         self._set_revack_pending(LOCAL, False)
+        self.log[LOCAL]['was_revoke_last'] = True
         # htlcs
         for htlc_id in self._maybe_active_htlc_ids[REMOTE]:
             ctns = self.log[REMOTE]['locked_in'][htlc_id]
@@ -215,7 +206,7 @@ class HTLCManager:
                 fee_update.ctn_local = self.ctn_latest(LOCAL) + 1
 
         # no need to keep local update raw msgs anymore, they have just been ACKed.
-        self.log['unacked_local_updates2'].pop(self.log[REMOTE]['ctn'], None)
+        self.log[LOCAL]['unacked_updates'].pop(self.log[REMOTE]['ctn'], None)
 
     @with_lock
     def _update_maybe_active_htlc_ids(self) -> None:
@@ -282,21 +273,26 @@ class HTLCManager:
         """We need to be able to replay unacknowledged updates we sent to the remote
         in case of disconnections. Hence, raw update and commitment_signed messages
         are stored temporarily (until they are acked)."""
-        # self.log['unacked_local_updates2'][ctn_idx] is a list of raw messages
+        # self.log[LOCAL]['unacked_updates'][ctn_idx] is a list of raw messages
         # containing some number of updates and then a single commitment_signed
         if is_commitment_signed:
             ctn_idx = self.ctn_latest(REMOTE)
         else:
             ctn_idx = self.ctn_latest(REMOTE) + 1
-        l = self.log['unacked_local_updates2'].get(ctn_idx, [])
+        l = self.log[LOCAL]['unacked_updates'].get(ctn_idx, [])
         l.append(raw_update_msg.hex())
-        self.log['unacked_local_updates2'][ctn_idx] = l
+        self.log[LOCAL]['unacked_updates'][ctn_idx] = l
 
     @with_lock
     def get_unacked_local_updates(self) -> Dict[int, Sequence[bytes]]:
-        #return self.log['unacked_local_updates2']
-        return {int(ctn): [bfh(msg) for msg in messages]
-                for ctn, messages in self.log['unacked_local_updates2'].items()}
+        #return self.log[LOCAL]['unacked_updates']
+        return {ctn: [bfh(msg) for msg in messages]
+                for ctn, messages in self.log[LOCAL]['unacked_updates'].items()}
+
+    @with_lock
+    def was_revoke_last(self) -> bool:
+        """Whether we sent a revoke_and_ack after the last commitment_signed we sent."""
+        return self.log[LOCAL].get('was_revoke_last') or False
 
     ##### Queries re HTLCs:
 
@@ -351,7 +347,6 @@ class HTLCManager:
             htlc_proposer: HTLCOwner,
             htlc_id: int,
     ) -> bool:
-        htlc_id = int(htlc_id)
         if htlc_id >= self.get_next_htlc_id(htlc_proposer):
             return False
         ctns = self.log[htlc_proposer]['locked_in'][htlc_id]
@@ -392,7 +387,6 @@ class HTLCManager:
             htlc_proposer: HTLCOwner,
             htlc_id: int,
     ) -> bool:
-        htlc_id = int(htlc_id)
         if htlc_id >= self.get_next_htlc_id(htlc_proposer):
             return False
         if htlc_id in self.log[htlc_proposer]['settles']:
@@ -495,8 +489,7 @@ class HTLCManager:
         return d
 
     @with_lock
-    def all_settled_htlcs_ever(self, subject: HTLCOwner, ctn: int = None) \
-            -> Sequence[Tuple[Direction, UpdateAddHtlc]]:
+    def all_settled_htlcs_ever(self, subject: HTLCOwner, ctn: int = None) -> Sequence[Tuple[Direction, UpdateAddHtlc]]:
         """Return the list of all HTLCs that have been ever settled in subject's
         ctx up to ctn.
         """
@@ -532,14 +525,16 @@ class HTLCManager:
         # sent htlcs
         for htlc_id in considered_sent_htlc_ids:
             ctns = self.log[whose]['settles'].get(htlc_id, None)
-            if ctns is None: continue
+            if ctns is None:
+                continue
             if ctns[ctx_owner] is not None and ctns[ctx_owner] <= ctn:
                 htlc = self.log[whose]['adds'][htlc_id]
                 balance -= htlc.amount_msat
         # recv htlcs
         for htlc_id in considered_recv_htlc_ids:
             ctns = self.log[-whose]['settles'].get(htlc_id, None)
-            if ctns is None: continue
+            if ctns is None:
+                continue
             if ctns[ctx_owner] is not None and ctns[ctx_owner] <= ctn:
                 htlc = self.log[-whose]['adds'][htlc_id]
                 balance += htlc.amount_msat
@@ -556,7 +551,8 @@ class HTLCManager:
         htlcs = []
         for htlc_id in considered_htlc_ids:
             ctns = self.log[htlc_proposer][log_action].get(htlc_id, None)
-            if ctns is None: continue
+            if ctns is None:
+                continue
             if ctns[ctx_owner] == ctn:
                 htlcs.append(self.log[htlc_proposer]['adds'][htlc_id])
         return htlcs
@@ -599,16 +595,16 @@ class HTLCManager:
         """Return feerate (sat/kw) used in subject's commitment txn at ctn."""
         ctn = max(0, ctn)  # FIXME rm this
         # only one party can update fees; use length of logs to figure out which:
-        assert not (len(self.log[LOCAL]['fee_updates']) > 1 and len(self.log[REMOTE]['fee_updates']) > 1)
+        assert not (len(self.log[LOCAL]['fee_updates']) > 0 and len(self.log[REMOTE]['fee_updates']) > 0)
         fee_log = self.log[LOCAL]['fee_updates']  # type: Sequence[FeeUpdate]
-        if len(self.log[REMOTE]['fee_updates']) > 1:
+        if len(self.log[REMOTE]['fee_updates']) > 0:
             fee_log = self.log[REMOTE]['fee_updates']
         # binary search
         left = 0
         right = len(fee_log)
         while True:
             i = (left + right) // 2
-            ctn_at_i = fee_log[i].ctn_local if subject==LOCAL else fee_log[i].ctn_remote
+            ctn_at_i = fee_log[i].ctn_local if subject == LOCAL else fee_log[i].ctn_remote
             if right - left <= 1:
                 break
             if ctn_at_i is None:  # Nones can only be on the right end
